-----------------------------------------------------------------------------
-- |
-- Module      :  GenC
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- Generate C code
--
-----------------------------------------------------------------------------

module GenC where

import qualified Core as C
import Code
import qualified Data.Map as M
import Type
import Data.Char
import Control.Monad
import Control.Monad.State
import System.IO
import Data.List
import Data.Maybe
import Debug.Trace

type TypeMap' = M.Map Id [Type]
data GenC = GenC { indentSize :: Int , typemap :: TypeMap' , nameMap :: M.Map String String , hh :: Handle , dump :: Bool , currentDistribution :: String }
               deriving (Eq, Show)

run defaults dumpCore fname typ likF initF sampF = do
  h <- openFile fname WriteMode
  let genC0   = GenC 0 (mkTypeMap' typ) M.empty h dumpCore ""
  evalStateT (gen fname defaults (mapReturnType likF) (map mapReturnType initF) (map mapReturnType sampF)) genC0
  hClose h
  return ()

makeGlobalMap :: C.Core -> M.Map String String
makeGlobalMap core = M.fromList [(g, renameV g) | g <- C.getGlobals core]

renameV [] = []
renameV (x:xs)
  | x `elem` okChars = x : renameV xs
  | otherwise        = '_' : renameV xs
  where okChars = ['a'..'z'] ++ ['A'..'Z'] ++ ['0'..'9'] ++ "_"

{- we need to get return types computed properly for the sampling
function.  we do the following:
 - if the variable type is Z or R, we return it directly
 - if the type is a vector *AND* it is a passed-in argument, we
   just overwrite and return void
 - if the type is a vector and is NOT a passed-in argument, we
   add an initial passed-in parameter that must be pre-allocated
   and post-freed
the key invariant this maintains is that any malloc that happens
in a function gets freed in the same function
-}

mapReturnType f =
  case (returnVal f, returnType f) of
    (Nothing,_)   -> f    -- no return: very easy
    (_,Type _ []) -> f    -- simple return value
    (Just (Var (V id [])),_)
      | id `elem` map fst (params f) -> f { returnVal = Nothing, returnType = tInt }
                                     -- ^-- passed in variable
    (Just (Var (V id [])),t) ->
                      f { returnVal = Nothing, returnType = tInt,
                          params = (id, t) : params f,
                          tempVars = filter ((/=id).fst) (tempVars f) }
    (Just expr, t) -> f { returnVal = Nothing, returnType = tInt,
                          params = ("return_value", t) : params f,
                          body = body f ++ [Assn (V "return_value" []) expr] }

type M a = StateT GenC IO a

--gen :: FilePath -> Function -> [Function] -> [Function] -> M ()
gen  fname (vals,loadD,loadM,numi) likF initF sampF = do
--  let initF' = filter (not.isDefined) initF
--  let sampF' = filter (not.isDefined) sampF

  let initF_sorted = let (a,b) = partition (("initialize_post_" `isPrefixOf`).name) initF in b++a

  statement ("/* COMPILE WITH: gcc -lm samplib.c stats.c " ++ fname ++ " */\n\n")

  statement "#include <stdio.h>\n#include <stdlib.h>\n#include <math.h>\n#include \"stats.h\"\n\n"
  statement "\n/**************************** SAMPLING ****************************/\n\n"
  sampF' <- mapM (\f -> genF f >>= \f' -> statement "\n" >> return f') sampF
  statement "\n/************************* INITIALIZATION *************************/\n\n"
  initF' <- mapM (\f -> genF f >>= \f' -> statement "\n" >> return f') initF_sorted
  statement "\n/*************************** LIKELIHOOD ***************************/\n\n"
  likF'  <- genF likF
  statement "\n/****************************** MAIN ******************************/\n\n"

  let initF'' = [f | f <- initF', not (isDefined (name f))]
  let sampF'' = [f | f <- sampF', not (isDefined (name f))]

  statement "int main() {\n"
  indent 1
  showIndent >> statement "double loglik,bestloglik;\n"
  showIndent >> statement "int iter;\n"

  tm <- (unTypeMap' . typemap) `liftM` get
  let vars = [(id,t) | (id,t) <- M.toList tm, not ("tmpSP" `isPrefixOf` id)]
  mapM_ (\ (id,t) -> showIndent >> genType t >> statement " " >> statement (renameV id) >> statement ";\n") vars

  let mallocDims = nub $ map (\ (_, Type _ ti) -> length ti-1) vars
  mapM_ (\i  -> showIndent >> genIT ("malloc_dim_" ++ show i, tInt) >> statement ";\n") [1..maximum (0:mallocDims)]

  statement "\n"

  showIndent >> statement "setall(time(0),time(0));   /* initialize random number generator */\n\n";

  mapM_ (\id -> showIndent >> statement (id ++ " = ???;    /* YOU MUST DEFINE THIS!!! */\n"))
        [renameV id | id <- M.keys tm, 
                      not ("tmpSP" `isPrefixOf` id), 
                      not (("initialize_" ++ id) `elem` map name initF'),
                      not (id `elem` commandDefined)]
  statement "\n"

  showIndent >> statement "/* variables defined with --define */\n"
  flip mapM_ vals $ \ (var,val) -> 
    showIndent >> statement (var ++ " = " ++ show val ++ ";\n")
  statement "\n"

  showIndent >> statement "/* variables defined with --loadD */\n"
  flip mapM_ loadD $ \ (var,(fn,high,dims,seps)) -> 
    showIndent >> statement (var ++ " = load_discrete" ++ (show (length dims)) ++ "(" ++
                                 "\"" ++ fn ++ "\"" ++ concatMap (\d -> ", &" ++ d) dims ++ ", &" ++ high ++ ");\n");
  statement "\n"

  showIndent >> statement "/* variables defined with --loadM */\n"
  flip mapM_ loadM $ \ (var,(fn,n,d)) -> 
    showIndent >> statement (var ++ " = load_matrix(\"" ++ fn ++ "\", &" ++ n ++ ", &" ++ d ++ ");\n");
  statement "\n"


--  mapM_ (\ (id,t) -> showIndent >> genType t >> statement " " >> statement (renameV id) >> statement ";     /* YOU MUST DEFINE THIS */\n") defs

  -- mallocs
  showIndent >> statement "fprintf(stderr, \"Allocating memory...\\n\");\n"
  mapM_ (\ (id,t) -> when (not $ null $ indices t) $ genMalloc 1 (statement id) t >> statement "\n") [(id,t) | (id,t) <- vars, not (id `elem` commandDefined)]
  statement "\n"

  showIndent >> statement "fprintf(stderr, \"Initializing variables...\\n\");\n"
  mapM_ (\f -> genFCall [] "" f) initF''

  statement "\n"
  showIndent >> statement ("for (iter=1; iter<=" ++ show numi ++ "; iter++) {\n")
  indent 1
  showIndent >> statement "fprintf(stderr, \"iter %d\", iter);\n"
  mapM_ (\f -> genFCall [] "" f) sampF''
  statement "\n"
  genFCall (filter ("post_" `isPrefixOf`) $ map fst vars) "loglik" likF'
  showIndent >> statement "fprintf(stderr, \"\\t%g\", loglik);\n"

  showIndent >> statement "if ((iter==1)||(loglik>bestloglik)) { bestloglik = loglik; fprintf(stderr, \" *\"); }\n"
  showIndent >> statement "fprintf(stderr, \"\\n\");\n"

  indent (-1)
  showIndent >> statement "}\n\n"

  mapM_ (\ (id,t) -> when (not $ null $ indices t) $ genFree 1 (statement id) t >> statement "\n") vars
  statement "\n"

  showIndent >> statement "return 0;\n"
  indent (-1)
  showIndent >> statement "}\n"
  where
    commandDefined = map fst vals ++ 
                     concatMap (\ (var,(_,high,dims,_)) -> var:high:dims) loadD ++ 
                     concatMap (\ (var,(_,n,d)) -> [var,n,d]) loadM
    isDefined x
      | "initialize_" `isPrefixOf` x = (drop 11 x) `elem` commandDefined
      | "resample_post" `isPrefixOf` x = True
      | "resample_"   `isPrefixOf` x = (drop  9 x) `elem` commandDefined
      | otherwise                    = False

genFCall postVars defStr f = genS (Call (name f) ret [Var (V id' []) | (id,_) <- params f
                                                                     , let id' = if ("post_" ++ id) `elem` postVars
                                                                                   then "post_" ++ id
                                                                                   else id])
  where
    ret = 
      case returnVal f of
        Nothing -> []
        Just e@(Var (V v [])) -> [e]
        _ -> [Var (V defStr [])]

genF f = do
  tm <- (unTypeMap' . typemap) `liftM` get
  let tm'  = foldr (\ (t,i) m -> M.insert t i m) tm (params f ++ tempVars f)
  let f'   = devectorizeF tm' $ liftVecF tm' $ f
  let tm'' = foldr (\ (t,i) m -> M.insert t i m) tm (params f' ++ tempVars f')
  let f''  = addDimsF tm'' f'
  let tm''' = foldr (\ (t,i) m -> M.insert t i m) tm (params f'' ++ tempVars f'')
  let newp = ((nub (codeIdsF f'') `intersect` M.keys tm''') \\ map fst (params f')) \\ map fst (tempVars f')
  let f'''  = f'' { params = params f'' ++ [(i,fromJust $ M.lookup i tm''') | i <- newp] }

  intermediate "initial function" (show f)
  intermediate "typemap" (show tm')
  intermediate "lifted" (show (liftVecF tm' f))
  intermediate "devectorized" (show f')
  intermediate "new parameters" (show newp)
  intermediate "almost final function" (show f'')
  intermediate "final function" (show f''')

  genF'  f'''
  return f'''

genF' f = do
  setDistribution (name f)
  showIndent
  statement $ case (returnVal f, returnType f) of
                (Nothing,_) -> "void "
                (_,Type T0Int  []) -> "int "
                (_,Type T0Real []) -> "double "
                (_,t) -> error ("GenC.genF: cannot generate functions with return type " ++ show t)
  statement $ name f ++ "("
  case params f of
    [] -> return ()
    (x:xs) -> genIT x >> mapM_ (\x -> statement ", " >> genIT x) xs
  statement ") {\n"
  indent 1

  -- create temporary variables
  let mallocDims = nub $ map (\ (_, Type _ ti) -> length ti-1) (nub $ tempVars f)
  let dvvVars    = nub $ [v | v <- codeIdsF f, "dvv_loop_var_" `isPrefixOf` v]
  mapM_ (\it -> showIndent >> genIT it >> statement ";\n") $ nub $ tempVars f
  mapM_ (\i  -> showIndent >> genIT ("malloc_dim_" ++ show i, tInt) >> statement ";\n") [1..maximum (0:mallocDims)]
  mapM_ (\i  -> showIndent >> genIT (i, tInt) >> statement ";\n") dvvVars

  mapM_ (uncurry defineType) $ params   f
  mapM_ (uncurry defineType) $ nub $ tempVars f
  mapM_ (\i -> defineType ("malloc_dim_" ++ show i) tInt) [1..maximum (0:mallocDims)]
  mapM_ (\i -> defineType i                         tInt) dvvVars

  -- malloc
  mapM_ (\ (id, t@(Type _ ti)) -> if null ti then return () else genMalloc 1 (statement id) t) $ nub $ tempVars f

  -- generate body
  tm <- (unTypeMap' . typemap) `liftM` get
  mapM_ genS $ body f

  -- free
  mapM_ (\ (id, t@(Type _ ti)) -> if null ti then return () else genFree   1 (statement id) t) $ nub $ tempVars f

  -- generate return
  case returnVal f of
    Nothing -> return ()
    Just e  -> showIndent >> statement "return (" >> genE e >> statement ");\n"
  indent (-1)
  showIndent
  statement "}\n"

  mapM_ (uncurry undefinedType) $ params   f
  mapM_ (uncurry undefinedType) $ nub $ tempVars f
  mapM_ (\i -> undefinedType ("malloc_dim_" ++ show i) tInt) [1..maximum (0:mallocDims)]
  mapM_ (\i -> undefinedType i                         tInt) dvvVars


genMalloc dim mkId t@(Type t0 ((C.Range _ lo hi):rs)) = do
  tm <- (unTypeMap' . typemap) `liftM` get
  let (lo', []) = codifyExpr tm lo
  let (hi', []) = codifyExpr tm hi
  showIndent
  mkId >> statement " = "
  statement "(" >> genType t >> statement ") "
  statement "malloc(sizeof(" >> genType (Type t0 rs) >> statement ") * (1+(" >> genE hi' >> statement ")-(" >> genE lo' >> statement ")));\n"
  when (not $ null rs) $ do
    let md = "malloc_dim_" ++ show dim
    setName ('#' : show dim) md
    showIndent
    statement $ "for(" ++ md ++ "="
    genE lo'
    statement $ "-1; " ++ md ++ "<"
    genE hi'
    statement $ "; " ++ md ++ "++) {\n"
    indent 1
    genMalloc (dim+1) (mkId >> statement ("[" ++md ++ "]")) (Type t0 rs)
    indent (-1)
    unsetName ('#' : show dim)
    showIndent
    statement "}\n"

genFree dim mkId t@(Type t0 ((C.Range _ lo hi):rs)) = do
  tm <- (unTypeMap' . typemap) `liftM` get
  let (lo', []) = codifyExpr tm lo
  let (hi', []) = codifyExpr tm hi
  when (not $ null rs) $ do
    let md = "malloc_dim_" ++ show dim
    setName ('#' : show dim) md
    showIndent
    statement $ "for(" ++ md ++ "="
    genE lo'
    statement $ "-1; " ++ md ++ "<"
    genE hi'
    statement $ "; " ++ md ++ "++) {\n"
    indent 1
    genFree (dim+1) (mkId >> statement ("[" ++md ++ "]")) (Type t0 rs)
    indent (-1)
    unsetName ('#' : show dim)
    showIndent
    statement "}\n"
  showIndent
  statement "free(" >> mkId >> statement ");\n"

genIT = uncurry genVarDef

genBType t0 = statement $ case t0 of { T0Int -> "int" ; T0Real -> "double" ; _ -> error ("GenC.genBType: unknown base type") }
genType (Type t0 ix) = do
  genBType t0
  statement $ replicate (length ix) '*'

genVarDef id t = do
  genType t
  statement $ ' ' : (renameV id)

genS (Loop id lo hi b) = do
  showIndent
  statement $ "for (" ++ (renameV id) ++ "="
  genE lo
  statement $ "-1; " ++ (renameV id) ++ "<"
  genE hi
  statement $ "; " ++ (renameV id) ++ "++) {\n"
  indent ( 1)
  mapM genS b
  indent (-1)
  showIndent
  statement "}\n"

genS (If c t e) = do
  showIndent
  statement $ "if ("
  genE c
  statement $ ") {"
  indent ( 1)
  mapM genS t
  indent (-1)
  case e of
    [] -> showIndent >> statement "}\n"
    _  -> do
      showIndent
      statement "} else {"
      indent ( 1)
      mapM genS e
      indent (-1)
      showIndent
      statement "}\n"

--genS (Assn v (Fun fn el)) | fn `elem` ["vec", "ID"] = genS (Call "vec" [] ((Var v):el))

genS (Assn v e) = do
  showIndent
  genV v
  statement " = "
  genE e
  statement ";\n";

genS (Incr v e) = do
  showIndent
  genV v
  statement " += "
  genE e
  statement ";\n";

genS (Call f [ret] arg)
  | "sample_" `isPrefixOf` f, multivarRange (drop 7 f) = do
  showIndent
  statement $ (renameV f) ++ "("
  genE ret
  mapM_ (\a -> statement ", " >> genE a) arg
  statement ");\n"

genS (Call f [] l) = do
  showIndent
  statement $ (renameV f) ++ "("
  when (not $ null l) $ let (a:as) = l in do
    genE a
    mapM_ (\a -> statement ", " >> genE a) as
  statement ");\n"

genS (Call f [Var v] l) = do
  showIndent
  genV v
  statement $ " = " ++ (renameV f) ++ "("
  when (not $ null l) $ let (a:as) = l in do
    genE a
    mapM_ (\a -> statement ", " >> genE a) as
  statement ");\n"

genS (Comment c) = showIndent >> statement ("/* " ++ c ++ " */\n")

genS s = error ("GenC.genS: cannot generate -- " ++ show s)


genE (Var v) = genV v
genE (Con (C.I i)) = statement (show i)
genE (Con (C.R r)) = statement (show r)
genE (Fun f [a,b]) 
  | f == ".+" || f == ".*" = do
  tm <- (unTypeMap' . typemap) `liftM` get
  let Type t1 t1i = inferCodeType tm a
  let Type t0 t0i = inferCodeType tm b

  if null t0i
    then genE (Fun (drop 1 f) [a,b])
    else do
      let f' = (if f == ".+" then "add_const_" else "mult_const_") ++
               (if t1 == T0Int then "i" else "r") ++ "_" ++
               (show (length t0i))
      statement f' >> statement "(" >> genE a >> statement ", " >> genE b
      mapM_ (\ (C.Range _ lo hi) ->
                 let (lo', []) = codifyExpr tm lo
                     (hi', []) = codifyExpr tm hi
                 in  statement ", (" >> genE hi' >> statement ")-(" >> genE lo' >> statement ")+1"
            ) (drop (length t1i) t0i)
      statement ")"

  | f `elem` ["=","~=","<=","<",">",">="] = do
  let f' = case f of { "=" -> "==" ; "~=" -> "!=" ; _ -> f }
  statement "((" >> genE a >> statement") "
  statement f'
  statement " (" >> genE b >> statement ") ? 1 : 0)"

  | all (not.isAlpha) f = do
  
  genOp f (statement "(" >> genE a >> statement ")") (statement "(" >> genE b >> statement ")")
  

genE (Fun f (e:es)) = do
  statement $ (renameV f) ++ "("
  genE e
  mapM_ (\e -> statement ", " >> genE e) es
  statement ")"

genV (V i el) = do
  nm <- nameMap `liftM` get
  let i' = M.findWithDefault i i nm
  when ((i == "") || (i == "?")) $ do
    curD <- currentDistribution `liftM` get
    fail ("GenC.genV: Type inference appears to have failed; try adding explicit type definitions for " ++ show curD)
  statement (renameV i') >> mapM_ (\e -> statement "[" >> genE e >> statement "]") el

statement s = get >>= \st -> lift (hPutStr (hh st) s)

intermediate heading str = do
  d <- dump `liftM` get
  when d (lift (hPutStrLn stderr heading >> hPutStrLn stderr str >> hPutStrLn stderr ""))

showIndent :: StateT GenC IO ()
showIndent = do
  i <- indentSize `liftM` get
  statement (replicate (2*i) ' ')
  return ()

setName   old new = get >>= \st -> put (st { nameMap = M.insert old new (nameMap st) })
unsetName old     = get >>= \st -> put (st { nameMap = M.delete old     (nameMap st) })

setDistribution d = get >>= \st -> put (st { currentDistribution = d })

--defineType    i t = get >>= \st -> put (st { typemap = M.insert i t (typemap st) })
--undefinedType i _ = get >>= \st -> put (st { typemap = M.delete i   (typemap st) })

defineType    i t = get >>= \st -> put (st { typemap = M.insertWith (++) i [t] (typemap st) })
undefinedType i _ = get >>= \st -> put (st { typemap = 
                                                 case M.lookup i (typemap st) of
                                                   Nothing     -> typemap st
                                                   Just []     -> M.delete i (typemap st)
                                                   Just [_]    -> M.delete i (typemap st)
                                                   Just (_:xs) -> M.insert i xs (typemap st) })


mkTypeMap' = M.map (:[])
unTypeMap' = M.map head


indent i = get >>= \st -> put (st { indentSize = indentSize st + i })

genOp ('.':f) a1 a2 = genOp f a1 a2
genOp "^" a1 a2 = statement "pow(" >> a1 >> statement ", " >> a2 >> statement ")"
genOp f a1 a2 = a1 >> statement (" " ++ f ++ " ") >> a2

-----------------------------------------------------------------------------

-- removes "sub" and "vec" and "ID"; also removes vector assignments and increments
devectorizeF tm f = f { returnVal = liftM     (devectorizeE tm') $ returnVal f
                      , body      = concatMap (devectorizeS tm') $ body f }
  where tm' = foldr (\ (v,t) -> M.insert v t) tm (params f ++ tempVars f)

devectorizeS tm (Loop i lo hi b) = [Loop i (devectorizeE tm lo) (devectorizeE tm hi) (concatMap (devectorizeS tm) b)]
devectorizeS tm (If c t e) = [If (devectorizeE tm c) (concatMap (devectorizeS tm) t) (concatMap (devectorizeS tm) e)]
--devectorizeS tm (Assn v (Fun "vec" l)) = devectorizeS tm $ devectorizeVec 1 v l
devectorizeS tm (Assn v e) =
  case inferCodeType tm (Var v') of
    Type _ [] -> [Assn v' e']
    Type T0Int  _ -> concatMap (devectorizeS tm) [Assn v' (Con (C.I 0)), Incr v' e']
    Type T0Real _ -> concatMap (devectorizeS tm) [Assn v' (Con (C.R 0)), Incr v' e']
  where v' = devectorizeV tm v ; e' = devectorizeE tm e
devectorizeS tm incr@(Incr v e)  =
--  | trace (show (incr, (inferCodeType tm (Var v'), inferCodeType tm e'))) True = 
  case (inferCodeType tm (Var v'), inferCodeType tm e') of
    (Type _ [], _)  -> [Incr v' e']
    (Type _ dV, Type _ dE) -> devectorizeS tm $ devectorizeIncr tm 1 v' e' dV dE
  where v' = devectorizeV tm v ; e' = devectorizeE tm e

devectorizeS tm (Call "vec" [] ((Var (V v sub)):val:dims)) = devectorizeS tm $ mkLoops 1 sub dims
  where
    mkLoops nn sub' [] = Assn (V v sub') val
    mkLoops nn sub' (lo:hi:dims) = Loop i lo hi [mkLoops (nn+1) (sub' ++ [Var (V i [])]) dims]
      where i = "dvv_loop_var_" ++ show nn

--devectorizeS tm c@(Call "vec" [] (ret:val:dims)) | trace (show (c, inferCodeType tm val)) False = undefined
{-
devectorizeS tm (Call "vec" [] (ret:val:dims)) =
  devectorizeS tm $
  case inferCodeType tm val of
    Type T0Int  _ -> Call ("vec_i_" ++ show ((length dims) `div` 2)) [] (ret : val : dims)
    Type T0Real _ -> Call ("vec_r_" ++ show ((length dims) `div` 2)) [] (ret : val : dims)
-}
--  where
--    makeDims (_:hi:xs) = hi : makeDims xs
--    makeDims [] = []

devectorizeS tm (Call v r a) = [Call v (map (devectorizeE tm) r) (map (devectorizeE tm) a)]
devectorizeS tm (Comment s) = [Comment s]

--devectorizeIncr tm num v e dV eV | trace (show (num,v,e,dV,eV)) False = undefined
devectorizeIncr tm num v e [] [] = Incr v e
devectorizeIncr tm num (V v vl) e ((C.Range _ lo hi):dV) eV@[] = 
  Loop i lo' hi' [devectorizeIncr tm (num+1) (V v (vl ++ [Var (V i [])])) e dV (drop 1 eV)]
  where 
    i = "dvv_loop_var_" ++ show num
    (lo', []) = codifyExpr tm lo
    (hi', []) = codifyExpr tm hi
devectorizeIncr tm num (V v vl) (Var (V e el)) ((C.Range _ lo hi):dV) ((C.Range _ lo2 hi2):eV) = 
  Loop i lo' hi' [devectorizeIncr tm (num+1) (V v (vl ++ [Var (V i [])])) (Var (V e (el ++ [Var (V i [])]))) dV eV]
  where 
    i = "dvv_loop_var_" ++ show num
    (lo', []) = codifyExpr tm lo 
    (hi', []) = codifyExpr tm hi
devectorizeIncr tm num (V v vl) e ((C.Range _ lo hi):dV) ((C.Range _ lo2 hi2):eV) = 
  Loop i lo' hi' [devectorizeIncr tm (num+1) (V v (vl ++ [Var (V i [])])) (devectorizeIncrE (Var (V i [])) e) dV eV]
  where 
    i = "dvv_loop_var_" ++ show num
    (lo', []) = codifyExpr tm lo 
    (hi', []) = codifyExpr tm hi
{-
devectorizeIncr tm num (V v vl) e [] ((C.Range _ lo hi):eV) = 
  Loop i lo' hi' [devectorizeIncr tm (num+1) (V v (vl ++ [Var (V i [])])) (devectorizeIncrE (Var (V i [])) e) [] eV]
  where 
    i = "dvv_loop_var_" ++ show num
    (lo', []) = codifyExpr tm lo 
    (hi', []) = codifyExpr tm hi
-}
devectorizeIncr tm num v e r1 r2 = error ("Code.devectorizeIncr: cannot handle: " ++ show (v,e,r1,r2))

devectorizeIncrE sub (Var (V i el)) = Var (V i (el ++ [sub]))
devectorizeIncrE sub (Con c) = error ("GenC.devectorizeIncrE: got constant")
devectorizeIncrE sub (Fun f [a,b])
  | f `elem` [".+",".-",".*","./"] = Fun f [a, devectorizeIncrE sub b]
  | f `elem` ["+","-","*","/","=","~=","<=","<",">",">=","||","&&"] = Fun f [devectorizeIncrE sub a, devectorizeIncrE sub b]
  | f == "^" = Fun f [devectorizeIncrE sub a, b]
devectorizeIncrE sub (Fun f (a:b:c:dims))
  | "mult_const_" `isPrefixOf` f ||
    "add_const_"  `isPrefixOf` f = Fun f (a:b:(devectorizeIncrE sub c):dims)
devectorizeIncrE sub (Fun f [a])
  | f `elem` ["log","exp"] = Fun f [devectorizeIncrE sub a]
devectorizeIncrE sub f@(Fun _ _) = error ("GenC.devectorizeIncrE: cannot devectorize " ++ show f ++ " with " ++ show sub)

devectorizeVec num v [val] = Assn v val
devectorizeVec num (V v sub) (val:lo:hi:dims) =
  Loop i lo hi [devectorizeVec (num+1) (V v (sub ++ [Var (V i [])])) (val:dims)]
  where i = "dvv_loop_var_" ++ show num

devectorizeE tm (Var v) = Var (devectorizeV tm v)
devectorizeE tm (Con c) = Con c
devectorizeE tm (Fun "sub" ((Var (V v el)):es)) = Var (V v (map (devectorizeE tm) (el++es)))
devectorizeE tm e@(Fun "vec" _) = error ("GenC.devectorizeE: vec not removed in: " ++ show e)
devectorizeE tm (Fun f (_:s:e:_))
  | "mult_const_" `isPrefixOf` f || "add_const_" `isPrefixOf` f
  , null $ indices $ inferCodeType tm e = devectorizeE tm (Fun "*" [s,e])

{- devectorizeE tm (Fun "vec" (val:dims)) = 
  case inferCodeType tm val of
    Type T0Int  _ -> Fun ("vec_i_" ++ show ((length dims) `div` 2)) (val : makeDims dims)
    Type T0Real _ -> Fun ("vec_r_" ++ show ((length dims) `div` 2)) (val : makeDims dims)
  where
    makeDims (_:hi:xs) = hi : makeDims xs
    makeDims [] = []
-}

devectorizeE tm (Fun f el) = Fun f (map (devectorizeE tm) el)

devectorizeV tm (V ('#':i) el) = V ("dvv_loop_var_" ++ i) el
devectorizeV tm (V v el) = V v (map (devectorizeE tm) el)

addDimsF tm f = f { body = map (addDimsS tm) $ body f }

addDimsS tm (Loop i lo hi b) = Loop i lo hi (map (addDimsS tm) b)
addDimsS tm (If c t e) = If c (map (addDimsS tm) t) (map (addDimsS tm) e)
addDimsS tm (Assn v e) = Assn v (addDimsE tm e)
addDimsS tm (Incr v e) = Incr v (addDimsE tm e)
addDimsS tm (Call f ret arg)
  | f `elem` ["sample_Mult", "sample_NorMV", "normalizeLog"] = Call f ret (map (addDimsE tm) arg ++ [lo',hi'])
  where 
    Type _ [C.Range _ lo hi] = inferCodeType tm (arg!!0)
    (lo', []) = codifyExpr tm lo
    (hi', []) = codifyExpr tm hi
addDimsS tm (Call f ret arg) = Call f ret (map (addDimsE tm) arg)
addDimsS tm (Comment c) = Comment c

addDimsE tm (Var (V id el)) = Var (V id $ map (addDimsE tm) el)
addDimsE tm (Con c) = Con c
addDimsE tm (Fun f arg)
  | f `elem` ["ldf_Mult", "ldf_NorMV", "sqrDiff"] = Fun f (map (addDimsE tm) arg ++ [lo',hi'])
  where 
    Type _ [C.Range _ lo hi] = inferCodeType tm (arg!!1)
    (lo', []) = codifyExpr tm lo
    (hi', []) = codifyExpr tm hi
addDimsE tm (Fun f arg) = Fun f $ map (addDimsE tm) arg
