----------------------------------------------------------------------------
-- |
-- Module      :  Marginalize
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- Implements marginalization (aka collapsing, ala Rao-Blackwellization)
--
-----------------------------------------------------------------------------

module Marginalize where

import Core
import Type

import Data.List
import Data.Maybe
import Control.Monad
import Debug.Trace
import qualified Data.Map as M
import Util


data Update = Update { indexVariable :: Id
                     , isIncr        :: Bool -- if false, then it's an assn
                     , numVars       :: Int
                     , variableRem   :: [SExpr] -> ([Range], Var, Expr)   -- (loop vars, update variable, incr/assn expression)
                     , variableAdd   :: [SExpr] -> ([Range], Var, Expr)
                     } 

marginalizeAll :: Core -> TypeMap -> [Id] -> (TypeMap, Core, [Id], [Update], [String])
marginalizeAll core0 tm0 coll =
  foldr (\id (tm,core,ids,upd,warn) ->
             case findMarginalizeType core id of
               Nothing -> (tm,core,ids,upd,("cannot figure out how to marginalize " ++ id):warn)
               Just f  ->
                 let (tm', id', core', upd') = f tm id core
                 in  (tm', core', id':ids, upd ++ upd', warn)
        ) (tm0,core0,[],[],[]) coll

findMarginalizeType (Core decls) var =
  case find ((==var).varId.fst3) decls of
    Just (_, Dist _ "Dir" _, _) -> Just marginalizeDirSym
    _ -> Nothing

-- marginalizeDirichlet v conj -> v', conj', update
--   v is the dirichlet-distributed variable whose dependents are
--   only ever multinomials, conj is the result of determinging
--   conjugacy.  we return a new variable name ("post_v")
--   which will need to be initialized, and a new core definition.
--   the new core differs from the original in the following ways:
--     1) the distribution over v has been removed
--     2) a definitional-distribution over post_v has been added
--     3) all references to v have been replaced by appropriate
--          references to post_v
--   finally, we also return a list of updates.  each update is
--   indexed by a variable that may change and, in doing so, will
--   affect the value of post_v.  each such update comes
--   with two functions: varRem and varAdd.  the first updates
--   post_v so that it *doesn't* depend on a the variable in
--   question; the second updates so that it *does* depend on the
--   variable in question.  these take as argument the sexprs to
--   the index variable.

marginalizeDirSym :: TypeMap -> Id -> Core -> (TypeMap, Id, Core, [Update])
marginalizeDirSym tm var core@(Core decls) = (tm', var', Core decls', updates)
  where
    globals = nub $ map (varId.fst3) decls

    varType@(Type _ typeIdx) = fromJust (M.lookup var tm)
    tm' = M.insert var' varType $ M.delete var tm

    var' = "post_" ++ var

    ([prior], nonPrior) = partition ((==var).varId.fst3) decls

    decls'  = variableDefinition : map (removeVariable var newE) nonPrior

    (variableDefinition, priorVal, varIdx, remainder) = 
      case prior of
        (V _ se, Dist _ "Dir" ((Func "+" [pv,rem]):el), rl) -> ((V var' se, Dist (V var' se) "Delta" (rem:el), rl), pv, se, rem)

    priorVal' =
      case priorVal of
        Func "vec" ((v@(Var _)):_) -> v
        Func "vec" ((c@(Con _)):_) -> c
        _ -> error ("Marginalize.marginalizeDirSym: priorVal does not appear symmetric: " ++ show priorVal)

    -- removeVariable replaces any occurance of (Var var sl) with (newE sl)
--    newE sl = Func ".+" [priorVal', Func ".*" [Func "/" [Con (R 1), Func "sum" [Var (V var' (take (length sl - 1) sl))]], Var (V var' sl)]]
    newE sl = Func ".+" [priorVal', Var (V var' sl)]

    -- find inner-most expression and associated ranges
    (rangeList, innerExpr) = traceInsideSums [] remainder

    -- the update var is anything that lives inside "innerExpr"
    updateVars = globals `intersect` getExprGlobals (map rangeIds rangeList) innerExpr
    updates    = map mkUpdate updateVars

    mkUpdate id =
      case nub $ findIdOcc id remainder of
        [V _ sl] -> Update id True (length sl) (mkStatement True sl) (mkStatement False sl)
        _ -> error ("Marginalize.marginalizeDirSym.mkUpdate: non-matching occurances of '" ++ id ++ "' found in " ++ show remainder)

    lu l x = case lookup x l' of { Nothing -> x ; Just y -> y }
      where l' = map (\ (Id v _, Id v' _) -> (v,v')) l

    mkStatement isRem sl0 sl = (map mkLoop loopVar, v0, addRem $ replaceVars (lu curVars) innerExpr)
      where 
        loopVar = mapMaybe (\ (v,t) -> case unqualifiedSE (mapMaybe sexprId sl0) v of { Nothing -> Nothing ; Just e -> Just (t,e) }) (zip varIdx typeIdx)
        _N = length sl0
        curVars = zip sl0 sl ++ map (\ (_,v) -> (Id v Nothing, Id v Nothing)) loopVar
        v0 = V var' (map (renameBy curVars) varIdx)  -- varIdx -- (take (length varIdx) sl)
        mkLoop (Range _ lo hi, id) = Range id lo hi
        addRem e = if isRem then Func ".-" [Con (R 0), e] else e
        
    traceInsideSums acc (SP False r@(Range v0 lo hi) e _) = traceInsideSums (r:acc) e
    traceInsideSums acc e = (acc, e)

    renameBy l x =
      case lookup x l of
        Nothing -> error ("Marginalize.renameBy: cannot find " ++ show x ++ " in " ++ show l)
        Just y  -> y

    replaceVars lu (Var (V id se)) = Var (V (lu id) (map (replaceVarsSE lu) se))
    replaceVars lu (Con c) = Con c
    replaceVars lu (Func f el) = Func f (map (replaceVars lu) el)
    replaceVars lu (SP b r e i) = SP b (replaceVarsR lu r) (replaceVars lu e) i
    replaceVarsSE lu (Id i c) = Id (lu i) c
    replaceVarsSE lu (Cn c) = Cn c
    replaceVarsR lu (Range i lo hi) = Range i (replaceVars lu lo) (replaceVars lu hi)

{-

    -- figure out which variables we need to loop over
    loopVar = mapMaybe (unqualifiedSE (map rangeIds rangeList)) varIdx

    varRem = C.Comment (show loopVar)
    varAdd = C.Comment ""
-}



containsVariable d v = v `elem` getDeclGlobals d

-- removeVariable assumes the variable in question never appears on the lhs nor in a range
removeVariable :: Id -> ([SExpr] -> Expr) -> Decl -> Decl
removeVariable id0 newE (var, dist, rl) = (var, removeVariableD id0 newE dist, rl)

removeVariableD id0 newE (Dist var id el) = Dist var id (map (removeVariableE id0 newE) el)
removeVariableD id0 newE (a :*: b) = removeVariableD id0 newE a :*: removeVariableD id0 newE b
removeVariableD id0 newE (a :^: e) = removeVariableD id0 newE a :^: removeVariableE id0 newE e
removeVariableD id0 newE (Prod r d id') = Prod r (removeVariableD id0 newE d) id'

removeVariableE id0 newE (Var (V id sl))
  | id == id0 = newE sl'
  | otherwise = Var (V id sl')
  where sl' = map (removeVariableSE id0 newE) sl
removeVariableE id0 newE (Func id el) = Func id (map (removeVariableE id0 newE) el)
removeVariableE id0 newE (SP b r e i) = SP b r (removeVariableE id0 newE e) i
removeVariableE id0 newE e0 = e0

removeVariableSE id0 newE se@(Id id c)
  | id == id0 = error ("Marginalize.removeVariableSE: found id " ++ show id0 ++ "in SExpr " ++ show se)
removeVariableSE id0 newE se = se

unqualifiedSE qual (Id i Nothing) | not (i `elem` qual) = Just i
unqualifiedSE qual (Id i c) | not (i `elem` qual) = error ("Marginalize.unqualifiedSE: qualified sexpr")
unqualifiedSE qual _ = Nothing

instance Show Update where
  showsPrec i (Update id isIncr numVars rem add) = 
    showString "Update { id=" . showString id . 
    showString ", isIncr=" . shows isIncr . 
    showString ", rem=" . showsPrec i (rem [Id ("var_" ++ show i) Nothing | i <- [1..numVars]]) . 
    showString ", add=" . showsPrec i (add [Id ("var_" ++ show i) Nothing | i <- [1..numVars]]) . 
    showString " }"

